#define RUNATONCE 3

//#define USEPGPLOT__CL
//#define DONTPERMORMIO
//#define TESTMODE

#include <fstream.h>
#include <strstream.h>

#include <gsl/gsl_math.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

#ifdef USEPGPLOT__CL
 #include <cpgplot.h>
#endif

#include "common.h"
#include "cl_array.h"
#include "cl_string.h"
#include "cl_time.h"

#define USEGSL__CL
#include "cl_hist.h"
#include "cl_partinfo.h"


#define PROGNAME "agesystem3"
#define VERSION "0.3"

#define OUTDIR "outer-"
#define INDIR "/inner-"
#define FILEN "file-"

/* global vars */
gsl_rng *r;  /* global generator */
clPartInfo* info;	/* global access struct */

/* for common.h */
int verbose=0;
int blocksize=BUFFERSIZE;
char *dummybuffer;

/* timer accounts */
#define N_diff 5
static const int n_timer=(int)(100/N_diff);
cl_Timer timer(n_timer);

/*for fs info*/
int fdsave=-1; 

/*for clsync*/
cl_String tmpdir="";

/* dist type */
int disttype=6;

/* to calc sizes */
int fixedlength=0;	
int min_file_size=0;
int mean_file_size=1024;
int min_mfile_size=1024;
int mean_mfile_size=16384;
int min_lfile_size=16384;
int mean_lfile_size=131072;

/* access directory */
int dirbase=100;
int dircounter=5000;

/* file size dist as histo */
cl_Histrand<unsigned long long, unsigned int> fs_dist(20,2,10);
cl_Histrand<unsigned long long, unsigned int> dtest(20,2,10);
/* dir dist as info */
clPartInforand* dinfo;	

/* find out directory to create in */
cl_String file(255);
cl_String fileadd(25);

unsigned long hash(const char *str){
	unsigned long l=strlen(str);
	const char *c=str;
	unsigned long sum = 0;
	for(unsigned long i=0;i<l;i++){
		unsigned long s=(unsigned long)*c++;
		sum+=(sum<<3)+s;
	}
	return sum;
}

void usage(){
	printf("Usage of %s v%s\n",PROGNAME,VERSION);
	printf(" -b blksize\t\t\tblocksize to use for buffered i/o\n");
	printf(" -c dir_base:dir_counter\tnumber of subdirectories to create\n");
	printf(" -d dirname\t\t\tchange to directory \"dirname\"\n");
	printf(" -f \t\t\t\tuse fixed length of bytes, given in blocksize\n");
	printf(" -h,-?\t\t\t\tthis text\n");
	printf(" -l string\t\t\tset file size creation options coded in \"string\"\n");
	printf("\t\t\t\tstring is a) 6:a:b:c:d:e:f ->\n\t\t\t\t a,b,c min_sizes; d,e,f max_sizes for distribution\n");
	printf("\t\t\t\t          b) 1:histfilename ->\n\t\t\t\t use distribution of measured histogram and use -c for dirs\n");
	printf("\t\t\t\t          c) 2:histdirname:histfilename ->\n\t\t\t\t use measured distribution for files and dirs\n");
	printf(" -n number_of_files\t\tmaximum number of files to create\n");
	printf(" -m mode\t\t\tset working mode to \"mode\"\n");
	printf("\t\t\t\t0 -> get info struct\n");
	printf("\t\t\t\t1 -> gauge only; use with -c and -l option \n");
	printf("\t\t\t\t2 -> age fs system; use with -c and -l option\n");
	printf("\t\t\t\t3 -> calc histograms; use with -l option\n");
	printf("\t\t\t\t4 -> create only directories given with -l oder -c option\n");
	printf(" -o access_mode\t\t\tset access mode to  \"access_mode\" \n\t    (eg. cw=%d, cws=%d, cwa=%d,cwas=%d, cwt=%d,cwts=%d)\n",O_CREAT|O_WRONLY,O_CREAT|O_WRONLY|O_SYNC,O_CREAT|O_WRONLY|O_APPEND,O_CREAT|O_WRONLY|O_APPEND|O_SYNC,O_CREAT|O_WRONLY|O_TRUNC,O_CREAT|O_WRONLY|O_TRUNC|O_SYNC);
	printf(" -p dirno:fileno\t\tstore maximum \"dirno\" dirs and \"fileno\" files in info struct\n");
	printf(" -r file\t\t\tload partition info from \"file\"\n");
	printf(" -s file\t\t\tsave partition info to \"file\"\n");
	printf(" -t\t\t\t\tsimply test input and exit without any action\n");
	printf(" -u min:max\t\t\tmin and max filesystem usage to work at\n");
	printf(" -v\t\t\t\tbe verbose\n");
	printf(" -y seed\t\t\tset random generator seed\n");
	printf(" -z\t\t\t\tsync with unmount (don't use or check/edit /root/bin/agesync.sh)\n");
	exit(1);
}

inline float fs_usage(){
	
	struct statfs buf_statfs;
	if(fstatfs(fdsave,&buf_statfs)==-1)
	{
		fprintf(stderr,"Error statfsing \"%s\", reason: %s\n",".",strerror(errno));
		exit(1);
	}
	return (float)(100-100.*(float)buf_statfs.f_bfree/(float)buf_statfs.f_blocks);
}

inline int f2d(int fn,int dircounter)
{
	return (fn % dircounter);
}

inline int d2d(int dn,int dirbase)
{
	return (dn % dirbase);
}

#ifdef firsttry__
inline int get_aktion(float f,float min=0,float max=100)
{	
	/*	1=delete
		2=create
	*/
	int fsu=(int)((f-min)/(max-min)*100);

	if(fsu>=98) return 1; /*delete for sure*/
	if(fsu<=2) return 2;  /*create for sure*/

	/*create number between 0 and 100 */
	int rannum=gsl_rng_uniform_int(r,101);
	
	if(rannum<fsu) return 1;
	else return 2;
}
#else
int rev=0;
inline int get_aktion(float f,float min=0,float max=100)
{	
	/*	1=delete
		2=create
	*/
	int fsu=(int)((f-min)/(max-min)*100);

	if(fsu>=95){
		rev=1;
		return 1;	/*delete for sure*/
	}
	if(fsu<=5){
		rev=0;
		return 2;	/*create for sure*/
	}
	/*create number between 0 and 100 */
	int rannum=gsl_rng_uniform_int(r,101);
	
	switch(rev){
		case 0:
			if(fsu<=75)
				if(rannum<=80) return 2; 
				else return 1;
			else if(fsu<=95)
				if(rannum<=70) return 2; 
				else return 1;
			break;
		case 1:
			if(fsu>=25)
				if(rannum>=20) return 1; 
				else return 2;
			else if(fsu>=5)
				if(rannum>=30) return 2; 
				else return 1;
		default:
			break;
	}

	if(rannum<fsu) return 1;
	else return 2;
}
#endif

void clsync(int flag=0)
{
#ifdef DONTPERMORMIO
	return;
#else
	sync();

	if(flag==0) return;
	if(tmpdir.length()==0) return;

	if(fdsave!=-1) close(fdsave);

	if(chdir("/root/bin")==0){
		cl_String cmd="agesync.sh ";
		cmd+=tmpdir;
		//cout << cmd << " ---- " << endl; 
		int ret=system((char *)cmd);
		if(ret!=0){
			fprintf(stderr,"Error calling /root/bin/agesync.sh\n");
			exit(1);
		}		
		chdir((char*)tmpdir);
	}

	fdsave=open(".",O_RDONLY);
	if(fdsave==-1){
		fprintf(stderr,"Error opening \"%s\", reason: %s\n",".",strerror(errno));
		exit(1);
	}		
#endif
}

int createdirs(){
	//create dirs
	cl_String dir(1024);
	int ret=-1;
	
	if(disttype==2){
		long d=dinfo->getnextdir(1);
		cl_String tmp(255);
		while(d!=-1){
			tmp=dinfo->getdir(d);
			char *stok=strtok((char*)tmp,"/");
			dir="";
			while(stok!=NULL)
			{
				if((strcmp(stok,"/")==0)||(strcmp(stok,".")==0)) {
					stok=strtok(NULL,"/");
					continue;
				}
				dir+=stok;
				dir+="/";
#ifndef DONTPERMORMIO
				ret=mkdir((char*)dir,S_IRUSR|S_IWUSR|S_IXUSR);
#endif
				//if(ret==-1) return -1;
				stok=strtok(NULL,"/");
			}
			ret=info->adddir((char*)dir);
			if(ret==-1){
				cerr << "Creating directory " << dir << " failed" << endl;
				return -1;
			}
			d=dinfo->getnextdir();
		}
	} else for(int i=0;i<dircounter;i++)
	{
		dir=OUTDIR;
		dir+=d2d(i,dirbase);
		dir+="/";
		mkdir((char*)dir,S_IRUSR|S_IWUSR|S_IXUSR);
		info->adddir((char*)dir);
		dir=OUTDIR;
		dir+=d2d(i,dirbase);
		dir+=INDIR;
		dir+=i;
		dir+="/";
#ifndef DONTPERMORMIO
		ret=mkdir((char*)dir,S_IRUSR|S_IWUSR|S_IXUSR);
#endif
		if(ret==-1) return -1;
		ret=info->adddir((char*)dir);
		if(ret==-1) return -1;
	}

	return 0;
}

int calclength()
{
	if(fixedlength==1) return(blocksize);

	int length=-1;
	
	if(disttype==6){
		int ran_size=gsl_rng_uniform_int(r,101);
		if(ran_size<90) length=(int)gsl_ran_gaussian_tail(r,min_file_size,mean_file_size); /*bytes*/
		else if(ran_size<=99) length=(int)gsl_ran_gaussian_tail(r,min_mfile_size,mean_mfile_size); /*bytes*/
		else length=(int)gsl_ran_gaussian_tail(r,min_lfile_size,mean_lfile_size); /*bytes*/
	} else if ((disttype==1) || (disttype==2)) {
		length=fs_dist.getrandnumber(r);
	}

	return length;
}

int calcdir(unsigned long long filecounter){ 
	int dirnumber=-1;

	if(disttype==2){
		dirnumber=dinfo->getrandir(r);
		file=dinfo->getdir(dirnumber);
	}
	else {
		dirnumber=f2d(filecounter,dircounter);
		file=OUTDIR;
		file+=d2d(dirnumber,dirbase);
		file+=INDIR;
		file+=dirnumber;
		file+="/";
	}



	file+=FILEN;
	file+=fileadd;
	file+=filecounter;

/*
	cl_String temp(50);
	temp=FILEN;
	temp+=fileadd;
	temp+=filecounter;
	unsigned long temp_nr=hash((char *)(temp));
	file+=temp_nr;
	file+=filecounter;
*/

	return -1;
}

int main(int argc,char **argv)
{
	int mode=-1;
	
	int flag=0;
	int testonly=0;
	int min_fsu=0;
	int max_fsu=75;
	long seed=0;

	unsigned long long filecounter=0;
	unsigned long long max_fileno=100000;
	unsigned long long countedbytes=0;
	unsigned long gmaxdirs=1000;
	unsigned long long gmaxfiles=100*gmaxdirs;

	int accessoptions=O_CREAT|O_WRONLY;

	cl_String rname(25);
	rname.clear();
	rname="";
	cl_String sname(25);
	sname.clear();
	sname="";
	cl_String lenstr(25);
	lenstr.clear();
	lenstr="";
	cl_String fdstr(25);
	fdstr.clear();
	fdstr="";
	cl_String ddstr(25);
	ddstr.clear();
	ddstr="";
		
	unsigned long long gfilecounter=0;
	unsigned long long gcountedbytes=0;
	unsigned long long gfilecounterdel=0;
	unsigned long long gcountedbytesdel=0;

	/*parse options*/
	int temp;
	char ch;
	while((ch=getopt(argc,argv,"b:c:d:fl:n:m:o:p:u:h?r:s:vty:z"))!=EOF){
		switch(ch)	{	
			case 'b':
				temp=atoi(optarg);
				if(temp<1)
				{
					fprintf(stderr, "Can't set blocksize to \"%s\".\n", optarg);
					usage();
				} else blocksize=temp;
				break;
			case 'c':{
				int i=0;
				char *opttok=strtok(optarg,":");
				while((opttok!=NULL)&&(i++<2)){
					//printf("%d-%s\n",++i,opttok);
					temp=atoi(opttok);
					if(temp<0)
					{
						fprintf(stderr, "Can't set directory counter to \"%s\".\n", optarg);
						usage();
					} 
					else if(i==1) dirbase=temp;
					else dircounter=temp;
					opttok=strtok(NULL,":");
				}
				break;}
			case 'f': fixedlength=1; break;
			case 'm': 
				temp=atoi(optarg);
				if(temp<-1)
				{
					fprintf(stderr, "Can't set mode to \"%s\".\n", optarg);
					usage();
				} else mode=temp;
				break;

			case 'n':
				temp=atol(optarg);
				if(temp<0)
				{
					fprintf(stderr, "Can't set maximum filenumber to \"%s\".\n", optarg);
					usage();
				} else max_fileno=temp;
				break;
			case 'l':
				lenstr=optarg;
				break;
			case 'o':
				temp=atoi(optarg);
				if(temp<1)
				{
					fprintf(stderr, "Can't set access options to \"%s\".\n", optarg);
					usage();
				} else accessoptions=temp;
				break;

			case 'p':{
				int i=0;
				char *opttok=strtok(optarg,":");
				while((opttok!=NULL)&&(i++<2)){
					//printf("%d-%s\n",++i,opttok);
					temp=atoi(opttok);
					if(temp<0)
					{
						fprintf(stderr, "Can't set directory counter to \"%s\".\n", optarg);
						usage();
					} 
					else if(i==1) gmaxdirs=temp;
					else { 
						gmaxfiles=temp;
						if(gmaxfiles<gmaxdirs) gmaxfiles=100*gmaxdirs;
					}
					opttok=strtok(NULL,":");
				}
				break;}

			case 'u':{
				int i=0;
				char *opttok=strtok(optarg,":");
				while((opttok!=NULL)&&(i++<3)){
					//printf("%d-%s\n",i,opttok);
					temp=atoi(opttok);
					if((temp<0)||(temp>100))
					{
						fprintf(stderr, "Can't set max usage to \"%s\".\n", optarg);
						usage();
					} else if(i==1) min_fsu=temp;
					else max_fsu=temp;
					opttok=strtok(NULL,":");
				}
				break;}
			case 'd':
				tmpdir=optarg;
				break;
			case 's':
				sname=optarg;
				break;
			case 'r':
				rname=optarg;
				break;
			case '?':
			case 'h': usage(); break;
			case 'v': verbose=1; break;
			case 't': testonly=1; break;
			case 'y': seed=atol(optarg); break;
			case 'z': flag=1; break;
			default: usage(); break;
		}
	}

	if(lenstr.length()>0){
		char *opttok=strtok((char*)lenstr,":");
		int i=0;
		int t=atoi(opttok);
		disttype=t;
		while((opttok!=NULL)&&(i++<t)){
			opttok=strtok(NULL,":");
			//printf("%d - %s\n",i,opttok);
			if(t==1){
				fdstr=opttok;
			} else if(t==2)
			{
				if(i==1) ddstr=opttok;
				else if(i==2) fdstr=opttok;

			} else if(t==6)
			{
				temp=atoi(opttok);
				if(temp<0)
				{
					fprintf(stderr, "Can't set file size length to \"%s\".\n",(char*)lenstr);
					usage();
				} 
				else if(i==1) min_file_size=temp;
				else if(i==2) min_mfile_size=temp;
				else if(i==3) min_lfile_size=temp;
				else if(i==4) mean_file_size=temp;
				else if(i==5) mean_mfile_size=temp;
				else if(i==6) mean_lfile_size=temp;
			}
		}
	}

	if(testonly==1){
		cout << "mode: " << mode << endl;

		cout << "accessoptions: " << accessoptions << endl;
		cout << "dirbase: " << dirbase << endl;
		cout << "dircounter: " << dircounter << endl;

		cout << "gmaxdirs: " << gmaxdirs << endl;
		cout << "gmaxfiles: " << gmaxfiles << endl;
		cout << "max_filenumber: " << max_fileno << endl;

		cout << "min_file_size: " << min_file_size << endl;
		cout << "mean_file_size: " << mean_file_size << endl;
		cout << "min_mfile_size: " << min_mfile_size << endl;
		cout << "mean_mfile_size: " << mean_mfile_size << endl;
		cout << "min_lfile_size: " << min_lfile_size << endl;
		cout << "mean_lfile_size: " << mean_lfile_size << endl;
		cout << "fixedlength: " << fixedlength << endl;

		cout << "min_fsu: " << min_fsu << endl;
		cout << "max_fsu: " << max_fsu << endl;

		cout << "tmpdir: " << tmpdir << endl;
		cout << "rname: " << rname << endl;
		cout << "sname: " << sname << endl;
		cout << "lenstr: " << lenstr << endl;
		cout << "ddstr: " << ddstr << endl;
		cout << "fdstr: " << fdstr << endl;

		cout << "blocksize: " << blocksize << endl;
		cout << "verbose: " << verbose << endl;
		cout << "flag: " << flag << endl;
		cout << "seed: " << seed << endl;
		exit(0);
	}

/* setup random number generator*/
//	gsl_rng_env_setup();
	gsl_rng_default_seed=seed;
	r=gsl_rng_alloc(gsl_rng_default);
		
	dummybuffer=new char[blocksize];
	memset(dummybuffer,31,blocksize);

	/* create partinfo */
	info=new clPartInfo(gmaxdirs,gmaxfiles);
	/* init timer */
	timer.initialize();

	ifstream ifs;
	if(rname.length()>0){
		ifs.open((char*)rname);
		if(mode>1000){
			mode-=1000;
			oldread(ifs,*info);
		} else ifs >> *info;

		ifs.close();
	}

	if((fdstr.length()>0)&&((mode!=0)&&(mode!=3))){
		ifs.open((char*)fdstr);
		ifs >> fs_dist;
		ifs.close();
	} 

	if((ddstr.length()>0)&&((mode!=0)&&(mode!=3))){
		dinfo=new clPartInforand(gmaxdirs,0);

		ifs.open((char*)ddstr);
		ifs >> *dinfo;
		ifs.close();
		dinfo->prepare();			
	} 
	
	ofstream ofs;
	if(sname.length()>0) ofs.open((char*)sname);

	if(mode!=3) if((tmpdir.length()>0)&&(chdir((char*)tmpdir)))
	{
		fprintf(stderr, "Can't change to directory \"%s\".\n", (char*)tmpdir);
		usage();
	}

/* open dir to get information about the file system*/
	fdsave=open(".",O_RDONLY);
	if(fdsave==-1){
		fprintf(stderr,"Error opening \"%s\", reason: %s\n",".",strerror(errno));
		exit(1);
	}		

	int ret=-1;
	int aktion=-1;
	int length=-1;
	int which_timer=0;
	float fsu=fs_usage();

	fileadd="t";
	{long t=(long)timer.get_cur_time();
	fileadd+=t;}
	fileadd+="-";

	switch(mode){
		case 0: /* do nothing */
			ret=info->walk_through_dirs(".");
			if(ret==-1){
				cerr << "Error walking through dirs!" << endl;
				cerr << *info << endl;
				exit(1);
			}
			if(sname.length()>0) ofs << *info << endl;			
			break;
		case 1: /* gauge only ala agesystem*/
			if(rname.length()==0){
				ret=createdirs();
				if(ret==-1){
					info->clear();
					ret=info->walk_through_dirs(".");
					if(ret==-1){
						cerr << "Error walking through dirs!" << endl;
						cerr << *info << endl;
						exit(1);
					}
				}
				clsync(flag);	/*commit changes*/
			}

			timer.timestamp();
			while((fsu=fs_usage())<max_fsu)
			{
				if(((int)fsu/(int)N_diff)!=which_timer){
					clsync(flag);
					timer.get_delta_t(which_timer,countedbytes);					
					timer.timestamp();
					which_timer=(int)fsu/(int)N_diff;
					countedbytes=0;
				}
				if(verbose==1) printf("Usage: %f with timer %d --- ",fsu,which_timer);

				if(max_fileno==0) filecounter++;
				else filecounter=(filecounter+1) % max_fileno;
						
				calcdir(filecounter);
				length=calclength();
#ifndef DONTPERMORMIO
				if(writefile((char *)file,length,accessoptions)==-1){
					fprintf(stderr,"Error creating file \"%s\" with length %d\n",(char *)file,length);
					exit(1);
				}
#endif
				countedbytes+=length;
				if(info->addfile((char*)file,length)==-1) {
					cerr << "Error adding file " << file << endl;
					cerr << *info << endl;
					exit(1);
				}
				if(verbose) printf("Created file \"%s\" with length %d\n",(char *)file,length);
			}	
			clsync(flag);

			timer.get_delta_t(which_timer,countedbytes);					

			if(sname.length()>0) ofs << *info << endl;			

			printf("\n\nTiming Results\n");
			for(int i=0;i<n_timer;i++) timer.print_stat(i);

			break;
		case 2: /* age only ala agesystem*/
			if(rname.length()==0){
				info->clear();
				ret=info->walk_through_dirs(".");
				if(ret==-1){
					cerr << "Error walking through dirs!" << endl;
					exit(1);
				}
			}

			timer.timestamp();
			which_timer=(int)fsu/(int)N_diff;
			while(filecounter<max_fileno)
			{
				fsu=fs_usage();				
				if((countedbytes>10000000)&&(((int)fsu/(int)N_diff)!=which_timer)){
					clsync(flag);
					timer.get_delta_t(which_timer,countedbytes);					
					timer.timestamp();
					which_timer=(int)fsu/(int)N_diff;
					countedbytes=0;
				}
				if(verbose==1) printf("Usage: %f with timer %d --- ",fsu,which_timer);

				aktion=get_aktion(fsu,min_fsu,max_fsu);
				
				for(int iii=0;iii<RUNATONCE;iii++) switch(aktion){
					case 1: /*delete*/
						while(1){
							long filetodel=gsl_rng_uniform_int(r,info->get_count_tot_files());
							length=info->getfile(filetodel);
							if(length>=0) {
#ifndef DONTPERMORMIO
								if(unlinkfile(info->getpath(filetodel))!=0){
									fprintf(stderr,"Error unlinking file \"%s\"\n",info->getpath(filetodel));
									exit(1);
								}
#endif
								gfilecounterdel++;
								gcountedbytesdel+=length;
								ret=info->delfile(filetodel);
								if(verbose==1) printf("Deleted file %ld with length %d\n",filetodel,length);
								break;
							}
						}
						break;
					case 2: /*create*/
						filecounter++;

						calcdir(filecounter);
						length=calclength();
#ifndef DONTPERMORMIO
						if(writefile((char *)file,length,accessoptions)==-1){
							fprintf(stderr,"Error creating file \"%s\" with length %d\n",(char *)file,length);
							exit(1);
						}
#endif
						countedbytes+=length;
						gfilecounter++;
						gcountedbytes+=length;
						if(info->addfile((char*)file,length)==-1) {
							cerr << "Error adding file " << file << endl;
							cerr << *info << endl;
							exit(1);
						}
						if(verbose) printf("Created file \"%s\" with length %d\n",(char *)file,length);
						break;
					default:
						fprintf(stderr, "Got unknown action: %d\n",aktion);
						exit(1);
				}
			}
			clsync(flag);
			timer.get_delta_t(which_timer,countedbytes);					

			if(sname.length()>0) ofs << *info << endl;			

			printf("\n\nTiming Results %llu %llu %llu %llu\n",gfilecounter,gfilecounterdel,gcountedbytes,gcountedbytesdel);
			for(int i=0;i<n_timer;i++) timer.print_stat(i);

			break;
		case 3: /* calc hists */
			if(info->check()==0){
				fprintf(stderr, "Found inconsistency in cl_partinfo::check()!\n");
				break;
			}
			ret=info->deldirs();
			if(verbose==1) printf("Deleted %d empty dirs!\n",ret);
			if(info->check()==0){
				fprintf(stderr, "Found inconsistency in cl_partinfo::check()!\n");
				break;
			}
			
			if(sname.length()>0) ofs << *info << endl;

			ret=info->getnextfile(1);
			if(ret!=1){
				while(ret!=-1)
				{
					fs_dist.add(info->getfile(ret));
					ret=info->getnextfile();
				}

				if(verbose) {
					cout << "Found fs_dist: " << endl;
					cout << fs_dist << endl;
				}
							
				if(fdstr.length()>0){
					ofstream ofs_;
					ofs_.open((char*)fdstr);
					ofs_ << fs_dist;
					ofs_.close();
				}
			}


			if(ddstr.length()>0){
				info->delfiles();
				//info->probdirs();
				info->check();
								
				ofstream ofs_;
				ofs_.open((char*)ddstr);
				ofs_ << *info << endl;
				ofs_.close();
			}

			break;

		case 4: /* create only dirs */
			ret=createdirs();
			if(ret==-1){
				cerr << "Error creating dirs!" << endl;
				if(ddstr.length()>0) {
					cerr << "dinfo: " << endl;
					cerr << *dinfo << endl;
				}
				exit(1);
			}
			clsync(flag);	/*commit changes*/

			break;
#ifdef USEPGPLOT__CL
		case 50: /* plot file histogram*/
			if(fdstr.length()==0){
				if(info->check()==0){
					fprintf(stderr, "Found inconsistency in cl_partinfo::check()!\n");
					break;
				}
				ret=info->deldirs();
				if(verbose==1) printf("Deleted %d empty dirs!\n",ret);
				if(info->check()==0){
					fprintf(stderr, "Found inconsistency in cl_partinfo::check()!\n");
					break;
				}
			
				ret=info->getnextfile(1);
				if(ret!=1){
					while(ret!=-1)
					{
						fs_dist.add(info->getfile(ret));
						ret=info->getnextfile();
					}

					if(verbose) {
						cout << "Found fs_dist: " << endl;
						cout << fs_dist << endl;
					}
				}	
			}
			
			if (cpgopen("?")<1) break;
			fs_dist.plot();
			cpglab("log_2 [KB]", "count %", "file size distribution (agesystem3)");
		    cpgclos();


#ifdef __first_try__
			{
			long cfiles=info->get_count_tot_files();
			float* farray=new float[cfiles];
			float pfl=0;
			float max=0;
			long i=0;
			long pf=info->getnextfile(1);
			while(pf!=-1){
				pfl=info->getfile(pf)/1024.;
				if(max<pfl) max=pfl;
				farray[i]=pfl;
				cout << i << ": " << pfl << " " << farray[i] << " " << max << endl;
				i++;
				pf=info->getnextfile();				
			}			
			if(i!=cfiles) break;
			if (cpgopen("?")<1) break;
			cpghist(cfiles,farray,0,max/10,100,0);
			//cpglab("value", "count %", "pgplot");
		    cpgclos();
			delete[]farray;
			}
#else
			
#endif
			break;
#endif		
#ifdef TESTMODE
		case 100: /* test */
			if(fdstr.length()==0) break;
			cout << (int)(10*fs_dist.getsum()) << endl;
			for(unsigned int i=0;i<10*fs_dist.getsum();i++)
			{
				unsigned int a=fs_dist.getrandnumber(r);
				//cout << a << endl;
				dtest.add(a);
			}
			//cout << fs_dist << endl;
			//cout << dtest << endl;
			//dinfo->set_verbose(1);			
			dinfo->prepare();			
			for(unsigned int i=0;i<10*fs_dist.getsum();i++){
				dinfo->getrandir(r);
			}
			//cout << *dinfo << endl;

			break;
#endif
		default:
			cout << "Info: " << endl;
			cout << *info << endl;
			if(fdstr.length()>0) {
				cout << "fs_dist: " << endl;
				cout << fs_dist << endl;
			}
			if(ddstr.length()>0) {
				cout << "dinfo: " << endl;
				cout << *dinfo << endl;
			}
			break;
	}

	if(sname.length()>0) ofs.close();
	if(fdsave!=-1) close(fdsave);
	if((ddstr.length()>0)&&((mode!=0)&&(mode!=3))) delete dinfo;
	delete info;
	delete[] dummybuffer;
	gsl_rng_free(r);
}
