/*

Copyright 2001 
by Constantin Loizides <constantin.loizides@isg.de>


This file is part of the journaling-filesystem fragmentation project.
You find the files belonging to that project in the Makefile
coming together with this file.

Source code concerning that project is free software; 
you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  
USA

*/

#ifndef cl_hist_h
#define cl_hist_h

#include <iostream.h>
#include <fstream.h>
#include <strstream.h>
#include <assert.h>
#include <math.h>

template <class T, class U>
class cl_Hist
{
	public:
		cl_Hist(unsigned int nr,int base=2,unsigned int start=0)
		{
			this->nr=nr;
			counter=new T[nr];
			x=new U[nr];
			s=T(0);
						
			long val=1;
			for(unsigned int i=0;i<start;i++) val*=base;
			for(unsigned int j=0;j<nr;j++){
				x[j]=(U)val;
				counter[j]=(T)0;
				val*=base;
			}
		}
		
		~cl_Hist()
		{ 
			delete[] counter;
			delete[] x;
		}

		T& getcounter(unsigned int i){
			assert(i<nr);
			return counter[i];
		}

		U& getx(unsigned int i){
			assert(i<nr);
			return x[i];
		}

		T& add(const U& val)
		{/*fill hist*/
			unsigned int i=0;
			while((i<nr-1)&&(val>x[i])) {
				i++;
				//cout << i << " " << x[i] << endl;
			}
			counter[i]++;
			s++;			
			return counter[i];
		}

		T& getsum()
		{
			return s;
		}
		
		friend ostream& operator<<(ostream &os,const cl_Hist &s)
		{
			for(unsigned int i=0;i<s.nr;i++) {
				os.width(3);
				os << i << ": ";
				os.width(10); 
				os << s.x[i] << " - " <<s.counter[i] << endl;
			}
			return os;
		}

		friend ofstream& operator<<(ofstream &ofs,const cl_Hist &s)
		{
			ofs << s.nr << endl;
			for(unsigned int i=0;i<s.nr;i++) {
				ofs << s.x[i] << " " <<s.counter[i] << endl;
			}
			return ofs;
		}
		
		friend ifstream& operator>>(ifstream &ifs,cl_Hist &s)
		{
			delete[] s.counter;
			delete[] s.x;
			s.s=(T)0;
			
			ifs >> s.nr;
			s.x=new U[s.nr];
			s.counter=new T[s.nr];
			for(unsigned int i=0;i<s.nr;i++) {
				ifs >> s.x[i] >> s.counter[i];
				s.s+=s.counter[i];
			}
			return ifs;
		}
#ifdef USEPGPLOT__CL
		void plot()
		{
			float max=0;
			float* y=new float[nr];
			float* x_=new float[nr];
			for(unsigned int i=0;i<nr;i++) {
				y[i]=(float)counter[i]/s;
				x_[i]=(float)i;
				if(max<y[i]) max=y[i];
				//cout << x_[i] << " " << y[i] << " " << max << endl;
			}
		
			cpgenv(0.,(float)nr,0.,max,0,0);
			cpgline(nr,x_,y);
		
			delete[] x_;
			delete[] y;
		}
#endif		
	protected:
		unsigned int nr;
		T *counter;
		T s;
		U *x;
};

template <class T,class U1,class U2>
class cl_Hist2
{
	public:
		cl_Hist2(unsigned int nr,int base=2,unsigned int start=0)
		{
			clear(nr,nr);
			xsetup(base,start);
			ysetup(base,start);
		}

		cl_Hist2(unsigned int nr1,unsigned int nr2)
		{
			clear(nr1,nr2);
		}

		cl_Hist2()
		{
			clear(100,100);
		}

		cl_Hist2(unsigned int nr1,const U1& von1,const U1& bis1,unsigned int nr2,const U2& von2,const U2& bis2,int flag=0)
		{
			clear(nr1,nr2);

			switch(flag){
				case 0:
					xlinsetup(von1,bis1);
					ylinsetup(von2,bis2);
					break;
				case 1:
					xsetup(von1,bis1);
					ylinsetup(von2,bis2);
					break;
				case 2:
					xlinsetup(von1,bis1);
					ysetup(von2,bis2);
					break;
			case 3:
					xsetup(von1,bis1);
					ysetup(von2,bis2);
					break;
			default:
					cout << "Error: Unknown flag: " << flag << endl;
					break;
			}	

		}
		
		~cl_Hist2()
		{ 
			delete[] counter;
			delete[] x;
			delete[] y;
		}

		T& getcounter(unsigned int i,unsigned j){
			assert(i<nr1);
			assert(j<nr2);
			return counter[i*nr2+j];
		}

		U1& getx(unsigned int i){
			assert(i<nr1);
			return x[i];
		}

		U2& gety(unsigned int j){
			assert(j<nr2);
			return y[j];
		}

		T& add(const U1& val1,const U1& val2)
		{/*fill hist*/
			unsigned int i=0;
			while((i<nr1-1)&&(val1>x[i])) {
				i++;
				//cout << i << " " << x[i] << endl;
			}
			unsigned int j=0;
			while((j<nr2-1)&&(val2>y[j])) {
				j++;
				//cout << j << " " << y[j] << endl;
			}
			counter[i*nr2+j]++;
			s++;			
			return counter[i*nr2+j];
		}

		T& getsum()
		{
			return s;
		}
		
		friend ostream& operator<<(ostream &os,const cl_Hist2 &s)
		{
			if(s.s==0) return os << "Sum equals zero" << endl;
			
			os << "          ";
			for(unsigned int j=0;j<s.nr2;j++) 
				if(s.yisempty(j)==1){os.width(10); os << s.y[j];}
			os << endl;

			for(unsigned int i=0;i<s.nr1;i++){
				if(s.xisempty(i)!=1) continue;
				for(unsigned int j=0;j<s.nr2;j++) {
					if(j==0) {os.width(10);os << s.x[i];}
					if(s.yisempty(j)!=1) continue;
					os.width(10); 
					os << s.counter[i*s.nr2+j];
				}
				os.width(10);
				os << s.sumx(i);
				os << endl;
			}
			os << "          ";
			for(unsigned int j=0;j<s.nr2;j++)
				if(s.yisempty(j)==1){
					os.width(10); 
					os << s.sumy(j);
				}
			os.width(10); 
			os << s.s << endl;

			return os;
		}
		friend ifstream& operator>>(ifstream &ifs,cl_Hist2 &s)
		{
			char buf[4096];
			unsigned int n=0;			
			unsigned int n1=0;			
			unsigned int n2=0;			
			T*  tmp_c=new T[s.nr];
			U1* tmp_x=new U1[s.nr1];						
			U2* tmp_y=new U2[s.nr2];						

			ifs.getline(buf,4096);
			char *stok=strtok(buf," ");
			while(stok!=NULL)
			{
				//cout << n2 << ": " << stok << endl; 
				assert(n2<s.nr2);
				istrstream(stok) >> tmp_y[n2];
				stok=strtok(NULL," ");
				n2++;
			}

			T* tmp_srow=new T[s.nr1];
			T* tmp_scol=new T[s.nr2];
			T* tmp=new T[n2+2];
			U1 tmp_U1;
			
			while(1){
				unsigned int nn=0;			
				ifs.getline(buf,4096);
				stok=strtok(buf," ");
				while(stok!=NULL)
				{
					assert(nn<n2+2);
					if(nn==0) istrstream(stok) >> tmp_U1;
					else istrstream(stok) >> tmp[nn];
					//cout << nn << ": " << stok << endl; 
					stok=strtok(NULL," ");
					nn++;
				}

				if(nn==n2+1){ //last row read
					T call=tmp[n2];
					T srow=(T)0;
					T scol=(T)0;
					tmp[0]=(T)tmp_U1;
					for(unsigned int i=0;i<n2;i++){
						tmp_scol[i]=tmp[i];
						scol+=tmp[i];
					}
					for(unsigned int i=0;i<n1;i++) srow+=tmp_srow[i];
					assert(srow==call);
					assert(scol==call);
					unsigned int nn=0;
					for(unsigned int i=0;i<n1;i++){
						srow=(T)0;
						for(unsigned int j=0;j<n2;j++){
							srow+=tmp_c[nn];
							nn++;
						}
						assert(srow==tmp_srow[i]);
					}
					for(unsigned int i=0;i<n2;i++){
						scol=(T)0;
						nn=i;
						for(unsigned int j=0;j<n1;j++){
							scol+=tmp_c[nn];
							nn+=n2;
						}
						assert(scol==tmp_scol[i]);
					}
								
					delete[] s.x;
					delete[] s.y;
					delete[] s.counter;
					s.counter=new T[n];
					s.x=new U1[n1];						
					s.y=new U2[n2];						
					s.s=call;				
					s.nr=n;
					s.nr1=n1;
					s.nr2=n2;
					for(unsigned int i=0;i<n1;i++) s.x[i]=tmp_x[i];
					for(unsigned int i=0;i<n2;i++) s.y[i]=tmp_y[i];
					for(unsigned int i=0;i<n;i++) s.counter[i]=tmp_c[i];
					
					break;
				}
				else assert(nn==n2+2);

				assert(n1<s.nr1);
				tmp_x[n1]=tmp_U1;
				tmp_srow[n1]=tmp[nn-1];
				n1++;

				for(unsigned int i=1;i<nn-1;i++)
				{
					assert(n<s.nr);
					tmp_c[n]=tmp[i];
					n++;
				}	
			}

			delete[] tmp;
			delete[] tmp_c;
			delete[] tmp_x;
			delete[] tmp_y;
			delete[] tmp_srow;
			delete[] tmp_scol;

			return ifs;
		}		
	protected:
		void clear(unsigned int nr1,unsigned int nr2)
		{

			this->nr1=nr1;
			this->nr2=nr2;
			this->nr=nr1*nr2;
			counter=new T[nr];
			x=new U1[nr1];
			y=new U2[nr2];
			s=T(0);
			
			for(unsigned int i=0;i<nr;i++) counter[i]=(T)0;
			for(unsigned int j=0;j<nr1;j++) x[j]=U1(0);
			for(unsigned int j=0;j<nr2;j++) y[j]=U2(0);
		}
		void xsetup(const U1& base1,const U1& start1)
		{
			U1 val=1;
			for(unsigned int j=0;j<nr1;j++){
				x[j]=start1+val;
				val*=base1;
			}
		}
		void ysetup(const U2& base2,const U2& start2)
		{
			U2 val=1;
			for(unsigned int j=0;j<nr2;j++){
				y[j]=start2+val;
				val*=base2;
			}
		}
		void xlinsetup(const U1& von1,const U1& bis1)
		{
			U1 steps1=(bis1-von1)/(nr1-1);
			for(unsigned int j=0;j<nr1;j++){
				x[j]=von1+j*steps1;
			}
		}
		void ylinsetup(const U2& von2,const U2& bis2)
		{
			U2 steps2=(bis2-von2)/(nr2-1);
			for(unsigned int j=0;j<nr2;j++){
				y[j]=von2+j*steps2;
			}
		}
		int xisempty(unsigned int i) const
		{
			assert(i<nr1);
			for(unsigned int j=0;j<nr2;j++) 
				if(counter[i*nr2+j]!=0) return 1;
			
			return 0;
		}
		int yisempty(unsigned int j) const
		{
			assert(j<nr2);
			for(unsigned int i=0;i<nr1;i++) 
				if(counter[i*nr2+j]!=0) return 1;
			
			return 0;
		}
		T sumx(unsigned int i) const
		{
			assert(i<nr1);			
			T ret=(T)0;
			for(unsigned int j=0;j<nr2;j++) 
				ret+=counter[i*nr2+j];
			return ret;
		}
		T sumy(unsigned int j) const
		{
			assert(j<nr2);			
			T ret=(T)0;
			for(unsigned int i=0;i<nr1;i++) 
				ret+=counter[i*nr2+j];
			return ret;
		}

		unsigned int nr;
		unsigned int nr1;
		unsigned int nr2;
		T *counter;
		T s;
		U1 *x;
		U2 *y;
};

#ifdef USEGSL__CL
template <class T, class U>
class cl_Histrand : public cl_Hist<T,U>
{
	public:
		cl_Histrand(unsigned int nr,int base=2,unsigned int start=0) : cl_Hist<T,U>(nr,base,start)
		{}
		U getrandnumber(const gsl_rng *r)
		{
			T ran=(T)gsl_rng_uniform_int(r,(unsigned int)s)+1;

			unsigned int i=0;
			U ss=counter[0];
			//cout << ran << "(" << s << "): " << i << " " << counter[i] << " " << ss << endl;
			while((i<nr-1)&&(ran>ss)) {
				i++;
				ss+=counter[i];
				//cout << ran << "(" << s << "): " << i << " " << counter[i] << " " << ss << endl;
			}

			double xlow=0;
			if(i!=0) xlow=x[i-1];
			double xhigh=x[i]-xlow;
			
			return (U)(xlow+gsl_rng_uniform(r)*xhigh);
		}
	protected:

};
#endif
#endif
