#ifndef clpartinfo_
#define clpartinfo_

#define DIRPREALLOC 35
#define FILEPREALLOC 15


struct fileinfo {
	char *name;
	long length;
	long dir;
};

struct dirinfo {
	char* path;
	long f_no;
	long c_no;
	long d_no;
};

#include <assert.h>
#include <string.h>
#include <iostream.h>
#include <fstream.h>
#include <sys/stat.h>
#include <errno.h>
#include <stdio.h>
#include <dirent.h>

class clPartInfo {
	public:
		clPartInfo(unsigned long d=1000,unsigned long f=100000){
			verbose=0;
			readflag=0;
			checkflag=0;
			set(d,f);
		}
		~clPartInfo(){
			free();
		}
		void reset(unsigned long d=1000,unsigned long f=100000){
			free();
			set(d,f);
		}
		void clear(){
			long tmpd=tot_dirs;
			long tmpf=tot_files;
			free();
			set(tmpd,tmpf);
		}
		void set_readflag(int f=0){	/*0=justcontent, 1=exact, 2=compdirs, 3=compress, 4=append */
			readflag=f;
		}
		void set_verbose(int v=0){
			verbose=v;
		}
		void set_checkflag(int c=0){
			checkflag=c;
		}
		long get_count_tot_dirs(){return count_tot_dirs;}
		long get_count_tot_files(){return count_tot_files;}
		long get_tot_dirs(){return tot_dirs;}
		long get_tot_files(){return tot_files;}
		int check(){
			long count_tot_dirs_=0;
			long count_tot_files_=0;

			long ret=getnextdir(1);
			while(ret!=-1) {
				if(pdirs[ret].f_no!=pdirs[ret].c_no-pdirs[ret].d_no){
					cerr << "Directory inconsistency: " << pdirs[ret].path << " " << pdirs[ret].f_no << " " << pdirs[ret].c_no << " " << pdirs[ret].d_no << endl;
					return 0;
				}
				count_tot_dirs_++;
				count_tot_files_+=pdirs[ret].f_no;
				if(checkflag>0){
					long f_=0;
					long ret2=getnextfile(1);
					while(ret2!=-1){
						if(pfiles[ret2].dir==ret) f_++;
						ret2=getnextfile();
					}
					if(f_!=pdirs[ret].f_no){
						cerr << "Directory + File inconsistency: " << pdirs[ret].path << " " << pdirs[ret].f_no<<" " << pdirs[ret].c_no << " " << pdirs[ret].d_no <<  " found " << f_ << endl;
						return 0;
					}
				}
				ret=getnextdir();
			}
			if(count_tot_dirs_!=count_tot_dirs){
				cerr << "Directory inconsistency: count_tot_dirs " << count_tot_dirs << " found " << count_tot_dirs_ << endl;
				return 0;
			}
			if(count_tot_files_!=count_tot_files){
				cerr << "Directory + File inconsistency: count_tot_files " << count_tot_files << " found " << count_tot_files_ << endl;
				return 0;
			}
			return 1;
						
		}
		long getnextdir(int reset=0){
			if(reset>0) dirpos=0;
			if(dirpos==tot_dirs) {
				dirpos=0; /* start from begin */
				return -1; 
			}
			
			long ndirpos=findnextdir(dirpos);
			if(ndirpos<dirpos){
				dirpos=0;
				return -1;
			}
			else{
				dirpos=ndirpos+1;
				return ndirpos;
			}
		}
		long getnextfile(int reset=0){
			if(reset>0) filepos=0;
			if(filepos==tot_files) {
				filepos=0; /* start from begin */
				return -1; 
			}
			
			long nfilepos=findnextfile(filepos);
			if(nfilepos<filepos){
				filepos=0;
				return -1;
			}
			else{
				filepos=nfilepos+1;
				return nfilepos;
			}
		}
		long adddir(const char *dirname){
			if(count_tot_dirs==tot_dirs) return -1;
			
			long w=finddir(dirname);
			if(w!=-1) return w;

			int l=strlen(dirname)+1;
			w=hash(dirname) % tot_dirs;
			w=getemptydir(w);
			if(w<0) return -1;

			if(l>DIRPREALLOC) {
				delete[] pdirs[w].path;
				pdirs[w].path=new char[l];
			}
			strcpy(pdirs[w].path,dirname);
			pdirs[w].f_no=0;
			pdirs[w].d_no=0;
			pdirs[w].c_no=0;

			count_tot_dirs++;

			return w;
		}
		long addfile(const char* pathname, long length=-1){
			int w=strlen(pathname);
			while((w>=0)&&(pathname[w]!='/')) w--;

			if((w==0)&&(pathname[w]!='/')) return addfile("/",pathname,length);

			loc_buf[w+1]='\0';
			strncpy(loc_buf,pathname,w+1);
			strcpy(loc_buf2,&pathname[w+1]);
			//cout << w << " " << loc_buf << " - " << loc_buf2 << endl;						

			return addfile(loc_buf,loc_buf2,length);
		}
		long addfile(const char* path, const char* name, long length=-1){
			if(count_tot_files==tot_files) return -1;

			long dir=adddir(path);
			if(dir<0) return -1;
			
			int l=strlen(name)+1;
			int l2=FILEPREALLOC;
			long w=count_tot_files;

			if(count_tot_free_files>0){
				count_tot_free_files--;
				w=pfreefiles[count_tot_free_files];
				l2=strlen(pfiles[w].name);
			}

			if(l>l2) {
				delete[] pfiles[w].name;
				pfiles[w].name=new char[l];
			}

			strcpy(pfiles[w].name,name);
			pfiles[w].length=length;
			pfiles[w].dir=dir;
			pdirs[dir].f_no++;
			pdirs[dir].c_no++;
			count_tot_files++;

			return w;
		}
		long getfile(long n) {
			assert(n>=0);
			if(n+1>tot_files) return -1;

			long dir=pfiles[n].dir;
			if(dir==-1) return -1;
			
			return pfiles[n].length;
		}
		long delfile(long n) {
			assert(n>=0);
			if(n+1>tot_files) return -1;

			long dir=pfiles[n].dir;
			if(dir==-1) return -1;

			pdirs[dir].f_no--;
			pdirs[dir].d_no++;
			pfiles[n].dir=-1;
			pfiles[n].length=-1;
			pfreefiles[count_tot_free_files]=n;
			count_tot_free_files++;
			count_tot_files--;
			return (count_tot_free_files-1);
		}
		void delfiles(long l=0) {
			for(int i=0;i<tot_files;i++){
				if(pfiles[i].length<l) continue;
				delfile(i);
			}				
		}
		long deldirs(int f=0) {
			long ret=getnextdir(1);
			long counterdel=0;
			while(ret!=-1) {
				if(pdirs[ret].f_no<=f){
					if(f!=0){
						long ret2=getnextfile(1);
						while(ret2!=-1){
							if(pfiles[ret2].dir==ret) delfile(ret2);
							ret2=getnextfile();
						}
					}
					pdirs[ret].f_no=-1;
					pdirs[ret].c_no=-1;
					pdirs[ret].d_no=-1;
					count_tot_dirs--;
					counterdel++;
				}
				ret=getnextdir(0);
			}
			return counterdel;
		}
		long getpath(long n,char** ptr) {
			assert(n>=0);
			if(n+1>tot_files) return -1;
			if(pfiles[n].dir==-1) return -1;
			
			unsigned int l=strlen(pdirs[pfiles[n].dir].path)+strlen(pfiles[n].name);

			if((*ptr==NULL)||(strlen(*ptr)<l)){
				delete[] *ptr;
				*ptr=new char[l+1];
			}
			// *ptr=new char[l+1];

			strcpy(*ptr,pdirs[pfiles[n].dir].path);
			strcat(*ptr,pfiles[n].name);
			return 0;
		}
		const char* getpath(long n) {
			assert(n>=0);
			if(n+1>tot_files) return NULL;
			if(pfiles[n].dir==-1) return NULL;
						
			strcpy(loc_buf,pdirs[pfiles[n].dir].path);
			strcat(loc_buf,pfiles[n].name);
			return loc_buf;
		}
		const char* getdir(long n) {
			assert(n>=0);
			if(n+1>tot_dirs) return NULL;
			if(pdirs[n].f_no==-1) return NULL;
						
			strcpy(loc_buf,pdirs[n].path);
			return loc_buf;
		}
		int walk_through_dirs(const char *path)
		{
			struct stat statbuf;
			if(lstat(path,&statbuf)==-1){
				fprintf(stderr,"Error stating \"%s\", reason: %s\n",path,strerror(errno));
				return -1;
			}
	
			if((S_ISLNK(statbuf.st_mode)==1)||(S_ISDIR(statbuf.st_mode)==0)) 
				return (stat_func(path,&statbuf)); /*not a dir*/
			
			int ret=stat_func(path,&statbuf); /*work on dir and save result*/

			DIR *dirz;
			dirent *direntz;
			if((dirz=opendir(path))==NULL) {
				fprintf(stderr,"Error opening dir \"%s\", reason: %s\n",path,strerror(errno));
				closedir(dirz);
				return (stat_func(path,&statbuf));
			}

			char npath[1024];
			while((direntz=readdir(dirz))!=NULL){
				/* ignore "." and ".." */
				if((strcmp(direntz->d_name,".")!=0)&&(strcmp(direntz->d_name,"..")!=0)){
					strcpy(npath,path);
					strcat(npath,"/");
					strcat(npath,direntz->d_name);
					strcpy(loc_buf,direntz->d_name);
					if(walk_through_dirs((char *)npath)!=0){
						ret=-1;
						break;
					}
				}
			}
	
			if(closedir(dirz)<0){
				fprintf(stderr,"Error closing dir \"%s\", reason: %s\n",(char *)path,strerror(errno));
				return -1;
			} 
			return ret;
		}
		friend int operator==(clPartInfo &s1,clPartInfo &s2){
			if(s1.count_tot_dirs!=s2.count_tot_dirs) return 0;
			if(s1.count_tot_files!=s2.count_tot_files) return 0;
			
			long runner=s1.getnextdir(1);
			long runner2=-1;
			while(runner!=-1){
				runner2=s2.finddir(s1.pdirs[runner].path);
				if(runner2==-1) return -1;
				if(s1.pdirs[runner].f_no!=s2.pdirs[runner2].f_no) return 0;
				if(s1.pdirs[runner].c_no!=s2.pdirs[runner2].c_no) return 0;
				if(s1.pdirs[runner].d_no!=s2.pdirs[runner2].d_no) return 0;
				//cout << s1.pdirs[runner].path << " " << s2.pdirs[runner2].path << endl;
				runner=s1.getnextdir();
			}
/* too slow without hash
			runner=s1.getnextfile(1);
			runner2=-1;
			while(runner!=-1){
				runner2=s2.findfile(s1.pfiles[runner].name);
				if(runner2==-1) return 0;
				if(s1.pfiles[runner].length!=s2.pfiles[runner2].length) return 0;
				//cout << s1.pfiles[runner].name << " " << s2.pfiles[runner2].name << endl;
				runner=s1.getnextfile();
			}
*/
			return 1;
		}
		friend int operator!=(clPartInfo &s1,clPartInfo &s2){
			if(s1==s2) return 0;
			return 1;
		}
		friend ostream& operator<<(ostream &os,const clPartInfo &s)
		{
			os << "Totdirs: " << s.count_tot_dirs << " (" << s.tot_dirs << ")" << endl;
			os << "Totfiles: " << s.count_tot_files << " (" << s.tot_files << ")" << endl;

			if(s.verbose>0){
				os << "Dirs: " << endl;
				for(int i=0;i<s.tot_dirs;i++) 
					if(s.pdirs[i].f_no>=0) os << i << ": " << s.pdirs[i].path << " " << s.pdirs[i].f_no << " " << s.pdirs[i].c_no << " " << s.pdirs[i].d_no << endl;
				os << "Files: " << endl;
				for(int i=0;i<s.tot_files;i++) 
					if(s.pfiles[i].dir>=0) os << i << ": " << s.pfiles[i].name << " " << s.pfiles[i].length << " " << s.pfiles[i].dir << endl;
			}
			return os;
		}
		friend int oldread(ifstream &ifs,clPartInfo &s)
		{
			long ctot_d,ctot_f;
			long tot_d,tot_f;
			long tmp_length;
			long tmp_f,tmp_c,tmp_d;
			char tmp_name[1024];
			char tmp_path[1024];
			int ret;

			ifs >> ctot_d >> tot_d;
			ifs >> ctot_f >> tot_f;

			switch(s.readflag){
				case 0:	//justcontent
					s.clear();
					break;
				case 1: //exact
					s.reset(tot_d,tot_f);
					break;
				case 2: //compdirs
					s.reset(ctot_d,tot_f);
					break;
				case 3:	//compress
					s.reset(ctot_d,ctot_f);
					break;
				case 4: //append
					break;
				default: break;
			}

			for(int i=0;i<ctot_f;i++){
				ifs >> tmp_name;
				ifs >> tmp_length;
				ifs >> tmp_path;
				ret=s.addfile(tmp_path,tmp_name,tmp_length);				
				if(ret==-1) {
					cerr << "Error while loading (oldread): " << tmp_path << " " << tmp_name << " " << tmp_length << endl; 
					exit(1);
					//return -1;
				}
			}

			for(int i=0;i<ctot_d;i++){
				ifs >> tmp_name;
				ifs >> tmp_f >> tmp_c >> tmp_d;
				ret=s.adddir(tmp_name);
				if((ret==-1)||(s.pdirs[ret].f_no!=tmp_f)) {
					cerr << "Error while loading (oldread): " << tmp_name << " " << tmp_f << " " << tmp_c << " " << tmp_d << endl; 
					exit(1);
					//return ret;
				}
				s.pdirs[ret].c_no=tmp_c;
				s.pdirs[ret].d_no=tmp_d;
			}
			
			return 0;
		}
		friend ifstream& operator>>(ifstream &ifs,clPartInfo &s)
		{
			long ctot_d,ctot_f;
			long tot_d,tot_f;
			long tmp_length;
			long tmp_f,tmp_c,tmp_d;
			char tmp_name[1024];
			char tmp_path[1024];
			char tmp_dumy[1024];
			int ret;

			ifs >> ctot_d >> tot_d;
			ifs >> ctot_f >> tot_f;

			switch(s.readflag){
				case 0:	//justcontent
					s.clear();
					break;
				case 1: //exact
					s.reset(tot_d,tot_f);
					break;
				case 2: //compdirs
					s.reset(ctot_d,tot_f);
					break;
				case 3:	//compress
					s.reset(ctot_d,ctot_f);
					break;
				case 4: //append
					break;
				default: break;
			}

			for(int i=0;i<ctot_f;i++){
				memset(tmp_name,'\0',1024);
				ifs >> tmp_dumy;
				while(strcmp(tmp_dumy,"/")!=0){
					strcat(tmp_name,tmp_dumy);
					ifs >> tmp_dumy;
				}
				ifs >> tmp_length;
				memset(tmp_path,'\0',1024);
				ifs >> tmp_dumy;
				while(strcmp(tmp_dumy,"/")!=0){
					strcat(tmp_path,tmp_dumy);
					ifs >> tmp_dumy;
				}
				ret=s.addfile(tmp_path,tmp_name,tmp_length);				
				if(ret==-1) {
					cerr << "Error while loading: " << tmp_path << " " << tmp_name << " " << tmp_length << endl; 
					exit(1);
				}
			}

			for(int i=0;i<ctot_d;i++){
				ifs >> tmp_dumy;
				memset(tmp_name,'\0',1024);
				while(strcmp(tmp_dumy,"/")!=0){
					strcat(tmp_name,tmp_dumy);
					ifs >> tmp_dumy;
				}
				ifs >> tmp_f >> tmp_c >> tmp_d;
				ret=s.adddir(tmp_name);
				if((ret==-1)||(s.pdirs[ret].f_no!=tmp_f)) {
					cerr << "Error while loading: " << tmp_name << " " << tmp_f << " " << tmp_c << " " << tmp_d << endl; 
					exit(1);
				}
				s.pdirs[ret].c_no=tmp_c;
				s.pdirs[ret].d_no=tmp_d;
			}
			
			return ifs;
		}
		friend ofstream& operator<<(ofstream &ofs,const clPartInfo &s)
		{
			ofs << s.count_tot_dirs << " " << s.tot_dirs << endl;
			ofs << s.count_tot_files << " " << s.tot_files << endl;

			for(int i=0;i<s.tot_files;i++) 
				if(s.pfiles[i].dir>=0) ofs << s.pfiles[i].name << " / " << s.pfiles[i].length << " " << s.pdirs[s.pfiles[i].dir].path << " / "<< endl;

			for(int i=0;i<s.tot_dirs;i++) 
				if(s.pdirs[i].f_no>=0) ofs << s.pdirs[i].path << " / " << s.pdirs[i].f_no << " " << s.pdirs[i].c_no << " " << s.pdirs[i].d_no << endl;

			return ofs;
		}
	protected:
		void set(unsigned long d=1000,unsigned long f=100000){
			count_tot_dirs=0;
			count_tot_files=0;
			count_tot_free_files=0;
			
			filepos=0;
			dirpos=0;

			tot_dirs=d;
			tot_files=f;

			pdirs=new dirinfo[tot_dirs];
			for(long i=0; i<tot_dirs; i++) {
				pdirs[i].path=new char[DIRPREALLOC];
				pdirs[i].f_no=-1;
				pdirs[i].c_no=-1;
				pdirs[i].d_no=-1;
			}
			pfiles=new fileinfo[tot_files];
			pfreefiles=new long[tot_files];
			for(long i=0; i<tot_files; i++) {
				pfiles[i].name=new char[FILEPREALLOC];
				pfiles[i].length=-1;
				pfiles[i].dir=-1;
				pfreefiles[i]=-1;	
			}
		}
		void free(){
			for(long i=0;i<tot_dirs;i++) delete[] pdirs[i].path;
			for(long i=0;i<tot_files;i++) delete pfiles[i].name;
			delete[] pdirs;
			delete[] pfiles;
			delete[] pfreefiles;
		}
		unsigned long hash(const char *str) const {
			unsigned long l=strlen(str);
			const char *c=str;
			unsigned long sum = 0;
			for(unsigned long i=0;i<l;i++){
				unsigned long s=(unsigned long)*c++;
				sum+=(sum<<3)+s;
			}
			return sum;
		}
		inline long getemptydir(long start) {
			assert(start>=0);
			if(count_tot_dirs==tot_dirs) return -1;
			if(start==tot_dirs) return -1;

			long ret=start;
			while(pdirs[ret].f_no!=-1) {
				ret++;
				if(ret>=tot_dirs) ret=ret % tot_dirs;
			}
			return ret;
		}		
		inline long findnextdir(long start) {
			assert(start>=0);
			if(count_tot_dirs==0) return -1;
			if(start==tot_dirs) return -1;

			long ret=start;
			while(pdirs[ret].f_no==-1) {
				ret++;
				if(ret>=tot_dirs) ret=ret % tot_dirs;
			}
			return ret;
		}		
		inline long finddir(const char *dirname){
			if(count_tot_dirs<1) return -1;

			int c=-1;
			int runner=0;
			long w=hash(dirname) % tot_dirs;

			while((pdirs[w].f_no!=-1)){
				c=strcmp(pdirs[w].path,dirname);
				if(c==0) break;
				w++;
				if(w>=tot_dirs) {
					w=w % tot_dirs;
					runner++;
					if(runner>=2) return -1;
				}
			}

			if(c==0) return w;
			else return -1;		
		}
		inline long findnextfile(long start) {
			assert(start>=0);
			if(count_tot_files==0) return -1;
			if(start==tot_files) return -1;

			long ret=start;
			while(pfiles[ret].dir==-1) {
				ret++;
				if(ret>=tot_files) ret=ret % tot_files;
			}
			return ret;
		}
		inline long findfile(const char *fname){
			if(count_tot_files<1) return -1;

			long runner=-1;
/*
			int c=-1;
			long w=hash(fname);
			long w2=-1;
			for(runner=0;runner<tot_files;runner++){
				if (pfiles[runner].dir==-1) continue;
				w2=hash(pfiles[runner].name);
				if(w!=w2) continue;
				c=strcmp(pfiles[runner].name,fname);
				if(c==0) break;
			}
*/
			runner=count_tot_files-1;
			int c=-1;
			while((runner>=0)&&((c=strcmp(pfiles[runner].name,fname)))!=0) runner--;

			if(c==0) return runner;
			else return -1;		
		}
		inline int stat_func(const char *path,const struct stat *sfs){
			if(S_ISDIR(sfs->st_mode)==1){
/*
				//printf("%s %s\n",path,loc_buf);
				
				if(count_tot_dirs==tot_dirs) return -1;
				int l=strlen(path)+1;
				if(l>DIRPREALLOC) {
					delete[] pdirs[count_tot_dirs].path;
					pdirs[count_tot_dirs].path=new char[l];
				}
				strcpy(pdirs[count_tot_dirs].path,path);
				pdirs[count_tot_dirs].f_no=0;
				pdirs[count_tot_dirs].d_no=0;
				pdirs[count_tot_dirs].c_no=0;
				count_tot_dirs++;
*/
				if(adddir(path)<0) {return -1;}
			} else if(S_ISREG(sfs->st_mode)==1){
				if(addfile(path,sfs->st_size)<0) {return -1;}
/*
				//printf("%s %s\n",path,loc_buf);
						
				if(count_tot_files==tot_files) return -1;

				int l=strlen(loc_buf)+1;
				if(l>FILEPREALLOC) {
					delete[] pfiles[count_tot_files].name;
					pfiles[count_tot_files].name=new char[l];
				}
				strcpy(pfiles[count_tot_files].name,loc_buf);
				pfiles[count_tot_files].length=sfs->st_size;

				int l2=strlen(path);
				memset(loc_buf,'\0',l2);
				strncpy(loc_buf,path,l2-l);
				l=count_tot_dirs-1;
				while((l>=0)&&(strcmp(pdirs[l].path,loc_buf))!=0) l--;
				pfiles[count_tot_files].dir=l;

				count_tot_files++;
*/
			} 
/*			else if(S_ISLNK(sfs->st_mode)==1){}*/

			return 0;
		};

		long count_tot_dirs;
		long count_tot_files;
		long count_tot_free_files;
		long tot_dirs;
		long tot_files;

		long dirpos;
		long filepos;

		fileinfo *pfiles;
		dirinfo *pdirs;
		long* pfreefiles;

		int verbose;
		int readflag;
		int checkflag;

		char loc_buf[1024];		
		char loc_buf2[1024];		
};

#ifdef USEGSL__CL
class clPartInforand : public clPartInfo 
{
	public:
		clPartInforand(unsigned long d=1000,unsigned long f=100000) : clPartInfo(d,f) 
		{
			readflag=3;
			sum=0;
			prepared=0;
		}
		long prepare()
		{
			if(prepared==1) return sum;

			delfiles();			
			long ret=getnextdir(1);
			while(ret!=-1) {
				sum+=pdirs[ret].c_no;
				pdirs[ret].f_no+=pdirs[ret].c_no;
				pdirs[ret].c_no=0;
				pdirs[ret].d_no=0;
				ret=getnextdir();
			}
			prepared=1;

			return sum;
		}
		long getrandir(const gsl_rng *r)
		{
			int ran=gsl_rng_uniform_int(r,sum)+1;

			long i=0;
			long ss=pdirs[0].f_no;
			//cout << ran << "(" << sum << "): " << i << " " << pdirs[i].f_no << " " << ss << endl;
			while((i<count_tot_dirs-1)&&(ran>ss)) {
				i++;
				ss+=pdirs[i].f_no;
				//cout << ran << "(" << sum << "): " << i << " " << pdirs[i].f_no << " " << ss << endl;
			}
			return i;
			
		}

	private:
		long sum;
		int prepared;
};

#endif

#endif
