/*

Copyright 2001 
by Constantin Loizides <constantin.loizides@isg.de>


This file is part of the journaling-filesystem fragmentation project.
You find the files belonging to that project in the Makefile
coming together with this file.

Source code concerning that project is free software; 
you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  
USA

*/


#ifndef cl_string_h
#define cl_string_h

#include "cl_array.h"
#ifndef HAVE_FORTRAN__
 #include <iostream.h>
#endif
#include <stdio.h>
#include <string.h>
#include <ctype.h>

class cl_String
{
	public:
		cl_String(unsigned long size=0) : data(size) 
		{
			clear();
		};
		
		cl_String(const cl_String &s) : data(0)
		{
			data.set_to((const char*)(s.data),s.data.length());
		};
		
		cl_String(const char* s) : data(0)
		{
			if(s) data.set_to(s,strlen(s)+1); //becauseof \O
			else clear();
		};
		
		~cl_String(void){}	
		
		void clear() { data.clear(); }
		
		unsigned long length(void)
		{ 
			return data.length()-1; 
		}
		
		char first(void)
		{
			return data[0];
		} 

		char last(void)
		{
			return data[data.length()-2];
		} 

		void cut(int von, int bis) {
		  if(von>0 && bis<=(signed)(data.length()) && von<bis) {
		    data.cut(von,bis);
		  };
		}

		cl_String mid(int von, int bis) 
		{
			cl_String erg=*(this);
			if(von>0 && bis<(signed)data.length() && von<=bis) 
			{
				erg.data.cut(bis,erg.data.length()-1);
		  		erg.data.cut(0,von-1);
			} else {erg="";};
		  	return erg;
		};

		int is_terminated()
		{
			return (data[data.length()-1]=='\0');
		}
		
		void to_lower()
		{
			for(unsigned int i=0;i<data.length();i++)
			{
				char c=(char)tolower((int)data.at(i));
				data[i]=c;
			}
		}

		void to_upper()
		{
			for(unsigned int i=0;i<data.length();i++)
			{
				char c=(char)toupper((int)data.at(i));
				data[i]=c;
			}
		}
		
		static cl_String toString(long num);
		static cl_String toString(long long num);
		static cl_String toString(unsigned long num);
		static cl_String toString(unsigned long long num);
		static cl_String toString(int num);
		static cl_String toString(unsigned int num);
		static cl_String toString(float num);
		static cl_String toString(double num);
		
		//operators
		cl_String& operator=(const char *s)
		{
			if(s) data.set_to(s,strlen(s)+1);
			else clear();
			return *this;
		};	

		cl_String& operator=(const cl_String &s)
		{
			data.set_to((const char*)(s.data),s.data.length());
			return *this;
		};	
		
		cl_String& operator+=(const char *s)
		{
			if(!s) return *this;
			data.cut_tail(1); //take care of \0
			data.append(s,strlen(s)+1);
			return *this;
		} 		
		
		cl_String& operator+=(const cl_String &s)
		{
			data.cut_tail(1); //take care of \0
			data+=s.data;
			return *this;
		}
		
		cl_String& operator+=(const char c)
		{
			data.cut_tail(1); //take care of \0
			data.append(&c,1);
			char d='\0';
			data.append(&d,1);
			return *this;
		}

		cl_String& operator+=(long l)
		{
			*this+=toString(l);
			return *this;
		}

		cl_String& operator+=(long long l)
		{
			*this+=toString(l);
			return *this;
		}

		cl_String& operator+=(unsigned long l)
		{
			*this+=toString(l);
			return *this;
		}

		cl_String& operator+=(unsigned long long l)
		{
			*this+=toString(l);
			return *this;
		}
		
		cl_String& operator+=(int i)
		{
			*this+=toString(i);
			return *this;
		}

		cl_String& operator+=(unsigned int i)
		{
			*this+=toString(i);
			return *this;
		}

		cl_String& operator+=(float i)
		{
			*this+=toString(i);
			return *this;
		}

		operator const char* () const
		{
			return (const char *)data;
		}

		operator const char* ()
		{
			return (const char *)data;
		}

		operator char* ()
		{
			return (char *)data;
		}
				
		char& operator[](int i)
		{
			if((i>=0)&&(i<(signed)(data.length()-1))) return data[i];
			return data[0];
		}		
		
		char& operator[](unsigned long i)
		{
			if((i>=0)&&(i<data.length()-1)) return data[i];
			return data[0];
		}		

		const char& operator[](int i) const
		{
			if((i>=0)&&(i<(signed)(data.length()-1))) return data[i];
			return data[0];
		}		
		
		const char& operator[](unsigned long i) const
		{
			if((i>=0)&&(i<data.length()-1)) return data[i];
			return data[0];
		}		

#ifndef HAVE_FORTRAN__
		friend ostream& operator<<(ostream &os, const cl_String &s);
		friend istream& operator>>(istream &is, cl_String &s);
#endif
		friend int operator==(const cl_String &s, const cl_String &t);
		friend int operator!=(const cl_String &s, const cl_String &t);
		friend int operator==(const cl_String &s, const char *t);
		friend int operator!=(const cl_String &s, const char *t);
				
	protected:
		cl_Array<char> data;
};

#endif



